/*
 * Copyright 1993-2008 NVIDIA Corporation.  All rights reserved.
 *
 * NOTICE TO USER:
 *
 * This source code is subject to NVIDIA ownership rights under U.S. and
 * international Copyright laws.  Users and possessors of this source code
 * are hereby granted a nonexclusive, royalty-free license to use this code
 * in individual and commercial software.
 *
 * NVIDIA MAKES NO REPRESENTATION ABOUT THE SUITABILITY OF THIS SOURCE
 * CODE FOR ANY PURPOSE.  IT IS PROVIDED "AS IS" WITHOUT EXPRESS OR
 * IMPLIED WARRANTY OF ANY KIND.  NVIDIA DISCLAIMS ALL WARRANTIES WITH
 * REGARD TO THIS SOURCE CODE, INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY, NONINFRINGEMENT, AND FITNESS FOR A PARTICULAR PURPOSE.
 * IN NO EVENT SHALL NVIDIA BE LIABLE FOR ANY SPECIAL, INDIRECT, INCIDENTAL,
 * OR CONSEQUENTIAL DAMAGES, OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS
 * OF USE, DATA OR PROFITS,  WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE
 * OR OTHER TORTIOUS ACTION,  ARISING OUT OF OR IN CONNECTION WITH THE USE
 * OR PERFORMANCE OF THIS SOURCE CODE.
 *
 * U.S. Government End Users.   This source code is a "commercial item" as
 * that term is defined at  48 C.F.R. 2.101 (OCT 1995), consisting  of
 * "commercial computer  software"  and "commercial computer software
 * documentation" as such terms are  used in 48 C.F.R. 12.212 (SEPT 1995)
 * and is provided to the U.S. Government only as a commercial end item.
 * Consistent with 48 C.F.R.12.212 and 48 C.F.R. 227.7202-1 through
 * 227.7202-4 (JUNE 1995), all U.S. Government End Users acquire the
 * source code with only those rights set forth herein.
 *
 * Any use of this source code in individual and commercial software must
 * include, in the user documentation and internal comments to the code,
 * the above Disclaimer and U.S. Government End Users Notice.
 */

#include <device_launch_parameters.h> // cuda keywords
#include <cuda_runtime.h> // cuda runtime functions

// includes, system
#include <stdio.h>
#include <stdlib.h>
#include <limits>

// Simple utility function to check for CUDA runtime errors
void checkCUDAError(const char *msg)
{
    cudaError_t err = cudaGetLastError();
    if( cudaSuccess != err)
    {
        fprintf(stderr, "Cuda error: %s: %s.\n", msg, cudaGetErrorString( err));
        exit(-1);
    }
}

int main( int argc, char** argv)
{
    const int array_length = 8;
    float *h_a; // pointer for host memory

    // pointers for device memory
    float *d_a, *d_b;

    // allocate and initialize host memory
    // on host: allocate an array of array_length elements consisting of array_length*sizeof(float) bytes
    h_a = new float[array_length];
    for ( unsigned int n = 0; n < array_length; n++ )
    {
        h_a[n] = static_cast<float>(n);
    }

    // **TODO ** exercise 1.a: allocate device memory for d_a and d_b

    // **TODO ** exercise 1.b: copy host array h_a to device array d_a

    // **TODO ** exercise 1.c: copy device array d_a to device array d_b

    // clear host memory
    for ( unsigned int n = 0; n < array_length; n++ )
    {
        h_a[n] = 0.0f;
    }

    // **TODO ** exercise 1.d: device to host copy

    // Check for any CUDA errors
    checkCUDAError("cudaMemcpy calls");

    bool matches = true;
    // verify the data on the host is correct
    for (unsigned int n = 0; n < array_length; n++ )
    {
        float diff = std::abs( h_a[n] - static_cast<float>(n) );
        matches = matches && ( diff < std::numeric_limits<float>::epsilon());
    }
    if ( !matches )
    {
        fprintf(stderr,"Error: Your Code is not correct!\n");
        exit(1);
    }

    // **TODO ** exercise 1.e: free device memory pointers d_a and d_b

    // Check for any CUDA errors
    checkCUDAError("cudaFree");

    // free host memory pointer h_a
    // Bonus: be sure to use cudaFreeHost for memory allocated with cudaMallocHost
    delete [] h_a;

    // If the program makes it this far, then the results are correct and
    // there are no run-time errors.  Good work!
    printf("Correct result!\n");

    return 0;
}

